
type value =
  | Num of int
  | Str of string;;

type fpnop = {
    op : value array -> fpnop list -> fpnop list;
    nvars : int;
    var: value list;
  };;

type instr = 
  | Const of value
  | Var of int
  | Call of string
  | Spec of fpnop list -> fpnop list  
  | Op of fpnop;;

type fpnfunc = {
    name : string;
    arity : int;
    code : instr list;
  };;

let functab = ref Hashtbl.create 250;;

let lookup_fn name = 
  let myfn = Hashtbl.find functab name in
  { op= execute_block myfn.code; nvars= myfn.arity; var= [] }
;;
  
let rec bind_calls = function
  | [] -> []
  | Call name :: insts -> Op (lookup_fn name) :: bind_calls insts
  | inst :: insts -> inst :: bind_calls insts
;;

let take_arg op v = { op= op.op; nvars= op.nvars - 1; var= v :: op.vars };;

let emit_val (op::ops) v =
  let new_op = take_arg op v in
  if new_op.nvars == 0
    then new_op.op (Array.of_list new_op.var) ops
    else new_op :: ops
;;

let rec execute_block block argl opstack = 
  match block with  
  | [] -> opstack
  | Const x :: insts -> execute_block insts argl (emit_val opstack x)
  | Var n :: insts -> execute_block insts argl (emit_val opstack argl.[n])
  | Spec spform :: insts -> execute_block insts argl (spform opstack)
  | Op o :: insts -> execute_block insts argl (o :: opstack)
;;
  